#include <iostream>
#include <fstream>

#include "tableau.h"
#include "formula.h"

using namespace std;

const Formula * livenessFor(std::string T, std::string C) {
	return
		new Release(
			new False(), 
			new Disjunction(
				new Negation(new AtomicFormula(T)),
				new Until(
					new True(),
					new AtomicFormula(C)
				)
			)
		);
}

Model * first_attempt(void) {
	Model & m = *(new Model());
	
	const Formula * c1 = new AtomicFormula("c1");
	const Formula * t1 = new AtomicFormula("t1");
	const Formula * c2 = new AtomicFormula("c2");
	const Formula * t2 = new AtomicFormula("t2");
	
	m.elementaryFormulas.insert(c1);
	m.elementaryFormulas.insert(t1);
	m.elementaryFormulas.insert(c2);
	m.elementaryFormulas.insert(t2);

	std::vector<State *> s(16, (State *) 0);
	for(unsigned i=0; i<16; ++i) {
		State * st = new State;
		
		if((i&1u)>0u) {
			st->insert(c2); //cerr << '1';
		}
	
		if((i&2u)>0u) {
			st->insert(t2); //cerr << '2';
		}

		if((i&4u)>0u) {
			st->insert(c1); //cerr << '4';
		}

		if((i&8u)>0u) {
			st->insert(t1); //cerr << '8';
		}
		
//		cerr << endl;

		s[i] = st;
		m.states.insert(st);
		m.transitionRelation[st] = new StateSet();
	}
	
	m.addTransition(s[0], s[8]);
	m.addTransition(s[0], s[2]);
	
	m.addTransition(s[8], s[4]);
	m.addTransition(s[8], s[10]);
	
	m.addTransition(s[2], s[1]);
	m.addTransition(s[2], s[10]);

	m.addTransition(s[4], s[0]);
	m.addTransition(s[4], s[6]);

	m.addTransition(s[10], s[6]);
	m.addTransition(s[10], s[9]);

	m.addTransition(s[1], s[0]);
	m.addTransition(s[1], s[9]);

	m.addTransition(s[6], s[2]);
	m.addTransition(s[6], s[5]);

	m.addTransition(s[9], s[8]);
	m.addTransition(s[9], s[5]);

	m.addTransition(s[5], s[4]);
	m.addTransition(s[5], s[1]);

	m.initialStates.insert(s[0]);
	
	return &m;
}

//==============================================================================
int main()
{
	const Formula *f = livenessFor("t1", "c1");
	std::string spec("AG(t1 -> AF c1)");

// 	const Formula *f = 
// 		new Conjunction(
// 			livenessFor("t1", "c1"),
// 			livenessFor("t2", "c2")
// 		);
// 			
// 	std::string spec("(AG(t1 -> AF c1)) & (AG(t2 -> AF c2))");


// 	const Formula *f = 
// 		new Conjunction(
// 			new Conjunction(
// 				livenessFor("t", "c"),
// 				new Disjunction(
// 					new NegatedAtomicFormula("c"),
// 					new NegatedAtomicFormula("t")
// 				)
// 			),
// 			
// 			new Conjunction(
// 					new NegatedAtomicFormula("n"),
// 					new NegatedAtomicFormula("c")
// 			)
// 		);				
// 			
// 	std::string spec("AG(t -> AF c) & (!c | !t)");



// 	Formula *f = 
// 			new Until(
// 				new True(),
// 				new AtomicFormula("c")
// 			);
// 	std::string spec("AF c");



// 	Formula * f = new Release(new False, 
// 		new Disjunction(
// 			new Negation(new AtomicFormula("c1")),
// 			new Negation(new AtomicFormula("c2"))
// 		));
// 	
// 	std::string spec("AG(!c1 | !c2)");
// 		
// 	std::cout << *f << std::endl; 
	Tableau t(f);
	t.removeUnreachable();
	t.removeNonTotalStates();
	
	cerr << t.states.size() << endl;
	
	ofstream smv("t.smv");
	t.toSMV(smv, spec);

	ofstream dot("t.dot");
	t.removeUnreachable();
	t.removeNonTotalStates();
	t.toDot(dot);

// 	Model * fa = first_attempt();
// 	ofstream fa_smv("fa.smv");
// 	fa->toSMV(fa_smv, std::string());
// 	fa->removeUnreachable();
// 	fa->removeNonTotalStates();
// 	
// 	ofstream fa_dot("fa.dot");
// 	fa->toDot(fa_dot);
// 	
// 	delete fa;	

//	std::cout << t.elementaryFormulas << std::endl;
//	std::cout << t.states << std:: endl;
//	std::cout << t.initialStates << std:: endl;


// 	for(StateSet::const_iterator i=t.states.begin(); i!=t.states.end(); ++i) {
// 		cout << "From: " << *i << std::endl;
// 		cout << "To: " << t.transition_relation[*i] << std::endl;
// 	}


// 	for(FormulaSet::const_iterator j=t.atoms.begin();
// 		j!=t.atoms.end();
// 		++j) {
// 		
// 		cout << *(*j) << endl;
// 	}
// 	cout << endl;
	
	return 0;
}
