#ifndef MODEL_H
#define MODEL_H

#include <set>
#include <map>
#include <iostream>
#include <string>
#include <vector>

//=============================================================================	
template<class T>
bool isInSet(const T & t, const std::set<T> & s) {
	std::set<T>::const_iterator ct = s.find(t);
	return ct!=s.end();
}
//==============================================================================
class Formula;
typedef std::set<const Formula *> FormulaSet;
std::ostream & operator<<(std::ostream & o, const FormulaSet & fs);
//==============================================================================
typedef FormulaSet State;
std::ostream & operator<<(std::ostream & o, const State & s);
typedef std::set<const State *> StateSet;
std::ostream & operator<<(std::ostream & o, const StateSet & s);
typedef std::map<const State *, StateSet *> TransitionRelation;
typedef std::set<StateSet *> FairnessConstraints;
//==============================================================================
class Model {
public:
	FormulaSet elementaryFormulas;
	StateSet states;
	StateSet initialStates;
	TransitionRelation transitionRelation;
	FairnessConstraints fairnessConstraints;

	bool isElementaryFormula(const Formula * f) const;
	StateSet statesHaving(const Formula * f) const;
	void toDot(std::ostream & o) const;
	void toSMV(std::ostream & o, std::string spec) const;
	void removeTransitionsTo(const State *s);
	void removeState(const State * s);
	void removeNonTotalStates(void);
	void addTransition(const State * from, const State * to) {
		transitionRelation[from]->insert(to);
	}

	StateSet computeReachable(void) const;
	void removeUnreachable(void);
	
	virtual ~Model();
	
	virtual std::string reformat(const Formula * f) const;
	
	Model * composeWith(const Model & m) const;
	
private:
	inline std::string stateFormula(const State * s, bool next) const;

protected:
	static bool increase(std::vector<unsigned> & v);

};
//==============================================================================
#endif
