#include <iostream>
using namespace std;

template<class T> 
class SmartPtr {
friend class SmartPtr<const T>;
private:
	T * ptr;

public:
	class MixIn {
	friend class SmartPtr<T>;
	friend class SmartPtr<const T>;
	private:
		mutable unsigned refCount;

	protected:
		inline MixIn(void) : refCount(0) {}
	};

	inline T * operator->() const { /*cerr << "->\n";*/ return ptr; }
	inline T & operator* () const { /*cerr << "*\n";*/ return *ptr; }
	
	inline SmartPtr(T * p) : ptr(p) { ++ptr->refCount; }
	inline SmartPtr(const SmartPtr & p) : ptr(p.ptr) { ++ptr->refCount; }
	inline ~SmartPtr() { if (--ptr->refCount==0) delete ptr; }
	inline SmartPtr & operator=(const SmartPtr & p) {
		++p.ptr->refCount;
		if (--ptr->refCount==0) delete ptr;
		ptr = p.ptr;
		return *this;
	}
	template<class C> inline bool operator==(const SmartPtr<C> & s) const {
		return ptr==s.ptr;
	}
};


template<class T> 
class SmartPtr<const T> {
friend class SmartPtr<T>;
private:
	const T * ptr;

public:
	inline const T * operator->() const { /*cerr << "->\n";*/ return ptr; }
	inline const T & operator* () const { /*cerr << "*\n";*/ return *ptr; }
	inline SmartPtr(const T * p) : ptr(p) { ++ptr->refCount; }
	template<class C> inline SmartPtr(const SmartPtr<C> & p) : ptr(p.ptr) { 
		++ptr->refCount; 
	}
	inline ~SmartPtr() { if (--ptr->refCount==0) delete ptr; }
	template<class C> inline SmartPtr & operator=(const SmartPtr<C> & p) {
		++p.ptr->refCount;
		if (--ptr->refCount==0) delete ptr;
		ptr = p.ptr;
		return *this;
	}
	template<class C> inline bool operator==(const SmartPtr<C> & s) const {
		return ptr==s.ptr;
	}
};

class A : public SmartPtr<A>::MixIn {
public:
	int a;
	A(void) : a(0) {}
	~A(void) { cerr << "destroyed\n"; }
};

int main() {
	SmartPtr<A> a(new A());
	SmartPtr<const A> b = a;
	SmartPtr<const A> c(a);
	c = a;
	if(a==b) { cerr << "equal\n"; }
	
	return 0;
}
